//funkce, ktere mohou pouzivat moduly
procedure StatorBusy(sMessage:string);         //naformatuje text do BusyDialogu
begin
  //Pokud byl pusteny nejaky externi modul (serie, odznacky, ...), muze byt
  //titulek prepsan. Proto je prepiseme svym
  GeoBusyCaption('Stator, '+_('generating statistics ...'));
  if(nMaxElem>0) then begin
    GeoBusyProgress(iRootElem,nMaxElem);
  end;
  GeoBusyKind(sMessage);
end;

procedure Output(str:string); //zapise string do vystupniho souboru
var s:string;
begin
  if(fOut=nil) then exit;
  s:=GlobalVarReplace(str);
  if(UpperCase(sOutCoding)<>'ANSI') then s:=UtfToCharset(AnsiToUtf(s),sOutcoding);
  if(fOut<>nil) then fOut.Write(s,Length(s));
end;

function ConvertToKm(fLength:extended; sUnitSrc:string):extended;
begin
  Result:=fLength;
  case sUnitSrc of
  'm',  'Meter': Result:=Result/1000.;
  'mi', 'Mile': Result:=Result/0.621;
  'yd', 'Yard': Result:=Result/1094;
  'ft', 'Foot': Result:=Result/3281;
  'in', 'Inch': Result:=Result/39370; 
  end;
end;

function ConvertFromKm(fLength:extended; sUnitTrg:string):extended;
begin
  Result:=fLength;
  case sUnitTrg of
  'm',  'Meter': Result:=Result*1000.;
  'mi', 'Mile': Result:=Result*0.621;
  'yd', 'Yard': Result:=Result*1094;
  'ft', 'Foot': Result:=Result*3281;
  'in', 'Inch': Result:=Result*39370; 
  end;
end;

function ConvertLength(fLength:extended; sUnitSrc,sUnitTrg:string):extended;
begin
  Result:=fLength;
  if(sUnitSrc=sUnitTrg) then exit;
  //prevedeme na kilometry
  Result:=ConvertToKm(Result,sUnitSrc);
  //prevedeme na pozadovane jednotky
  Result:=ConvertFromKm(Result,sUnitTrg);
end;

function GetLengthAbbr(sUnit:string):string;
begin
  Result:=sUnit;
  case sUnit of
  'Kilometer': Result:='km';
  'Meter': Result:='m';
  'Mile': Result:='mi';
  'Yard': Result:='yd';
  'Foot': Result:='ft';
  'Inch': Result:='in'; 
  end;
  Result:=LowerCase(Result);
end;

//nacte kroky do pole a vrati pocet nactenych/vyuzitych kroku
function ReadStepsToArray(str:string; arr:array of integer):integer;
var s:string;
begin
  Result:=0;
  while(str<>'') do begin
    if(Result>high(arr)) then break;
    s:=Fetch(str,',');
    arr[Result]:=StrToInt(s);
    Inc(Result);
  end;
end;

function EncodeDataExtended(iNum:integer):string;
// prevede cislo v rozsahu 0-4095 na Extended Encoding Format Google Image Charts (kiReGPS)
var
  sEncode:string;
begin
  sEncode:='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-.';
  Result:=Copy(sEncode,(iNum div 64)+1,1)+Copy(sEncode,(iNum mod 64)+1,1);
  if (iNum > 4095)or(iNum < 0) then Result := '__';
end;

function EncodeNormDataExtended(iNum,iMax:integer):string;
// normalizuje slo na maximum 4095 a pevede na Extended Encoding Format Google Image Chart (kiReGPS)
begin
  if iMax>0 then Result:=EncodeDataExtended(Round(4095.0*iNum/iMax)) else Result:=EncodeDataExtended(0);
end;

function EncodeDataSimple(iNum:integer):string;
// prevede cislo v rozsahu 0-61 na Simple Encoding Format Google Image Charts (kiReGPS)
var
  sEncode:string;
begin
  sEncode:='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
  Result:=Copy(sEncode,(iNum mod 62)+1,1);
  if (iNum > 61)or(iNum < 0) then Result := '_';
end;

function EncodeNormDataSimple(iNum,iMax:integer):string;
// normalizuje slo na maximum 61 a pevede na Simple Encoding Format Google Image Chart (kiReGPS)
begin
  if iMax>0 then Result:=EncodeDataSimple(Round(61.0*iNum/iMax)) else Result:=EncodeDataSimple(0);
end;

function EncodeGoogleLabel(str:string):string;
begin
  //prevede vstupni kodovani do UTF a to zakoduje pro pouziti v URL
  //vystup je vzdy v UTF, protoze Google to vyzaduje
  //if(UpperCase(sOutCoding)<>'ANSI') then str:=UtfToCharset(AnsiToUtf(str),sOutcoding);
  str:=AnsiToUtf(str);
  Result:=EncodeUrlElement(str);
  //Result:=str;
end;

function GraphAxisStep(eRange:extended):integer;
// automatick nastaven kroku osy grafu 
begin
  // Pascal script zejm nezn exponenciln funkci, proto budeme naivn testovat
  Result:=1;
  if eRange>10 then Result:=2;
  if eRange>25 then Result:=5;
  if eRange>50 then Result:=10;
  if eRange>100 then Result:=20;
  if eRange>250 then Result:=50;
  if eRange>500 then Result:=100;
  if eRange>1000 then Result:=200;
  if eRange>2500 then Result:=500;
  if eRange>5000 then Result:=1000;
  if eRange>10000 then Result:=2000;
  if eRange>25000 then Result:=5000;
  if eRange>50000 then Result:=10000;
  if eRange>100000 then Result:=20000;
  if eRange>250000 then Result:=50000;
  if eRange>500000 then Result:=100000;
end;

function FormatFloatDot(const sFormat:string; eValue:extended):string;
// naformtuje slo s desetinnou tekou (nezvisle na systmov definici oddlovae)
// (vhodn pro Google Image Chart, kde je nutn desetinn oddlova teka) 
begin
  Result:=ReplaceString(FormatFloat(sFormat,eValue),',','.');
end;

function SetDefaultColumnAlign(sName:string;bHeader:boolean):string; //vrati implicitni zarovnani podle jmena sloupce
begin
  //vsechny sloupce jsou implicitne zarovnany vlevo, jen vybrane zarovname jinak
  //u ciselnych sloupcu je jinak zarovnana hlavicka
  Result:='left';
  case UpperCase(sName) of
    'TYPE ICON':            Result:='center';
    'SIZE ICON':            Result:='center';
    'DIFFICULTY':           Result:='center';
    'DIFFICULTY ICON':      Result:='center';
    'TERRAIN':              Result:='center';
    'TERRAIN ICON':         Result:='center';
    //'IDTAG':              Result:='center';
    'STATUS ICON':          Result:='center';
    'DATECREATE':           Result:='center'; 
    'DISTANCE':             Result:='right';
    'DIRECTION':            Result:='right';
    'DIRECTION ICON':       Result:='center';
    'COUNTRY FLAG':         Result:='center';
    'FOUND DATE':           Result:='center';
    'FOUND TIME':           Result:='center';
    'FOUND DATETIME':       Result:='center';
    'DATE OF LAST FOUND':   Result:='center';
    'COUNT':                if(bHeader) then Result:='center' else Result:='right';
    'ORDER FINDING',        //toto je spatne a nechavan to jen z historickych duvodu
                            //aby neprestaly chodit stare selecty
    'ORDER BY FINDS':       if(bHeader) then Result:='center' else Result:='right';
    'ORDER ROW':            if(bHeader) then Result:='center' else Result:='right';
    'MILESTONE':            if(bHeader) then Result:='center' else Result:='right';
    'MILESTONE, INTERVAL':  if(bHeader) then Result:='center' else Result:='right';
    'PERCENT':              if(bHeader) then Result:='center' else Result:='right';
    'PERCENT EXPLICIT':     if(bHeader) then Result:='center' else Result:='right';
  end;
end;

function OutputTableListHeader(sCols:string):string;
var sCol,s1,s2,sWidth:string;
    flags:REC_COLUMN_FLAGS;
    //ini:TIniFile;
begin
  //protoze sirka muze byt dana i primo v px, musime nahradit styl a v nem sirku,
  //jinak pri globalni nahrade je pouzita sirka cele statistiky
  Result:='<table style="'+GlobalVarGet('StyleTableList')+'"><thead>'+CRLF+'  <tr style="%StyleTableListTr%">'+CRLF;
  sWidth:=IntToStr(iWidth);
  Result:=ReplaceString(Result,'%Width%',sWidth);
  while (sCols<>'') do begin
    s1:=Fetch(sCols,'|');
    ParseColumnFlags(s1,flags);
    s2:=Fetch(s1,'='); //odpareme pripadny priznak zarovnani
    if(flags.sAlign='hide') then continue; //preskocime neviditelne sloupce
    if(flags.sAlign='') then flags.sAlign:=LowerCase(SetDefaultColumnAlign(flags.sName,true));
    sCol:=flags.sName; //odpareme pripadny komentar
    //odpareme pripadnou hodnotu tagu
    if(Pos('tag-',sCol)=1) then begin
      s1:=sCol;
      sCol:=Fetch(s1,'#');
    end;
    sWidth:='';
    if(UpperCase(sCol)='HISTOGRAM') or (UpperCase(sCol)='HISTOGRAM EXPLICIT') then sWidth:=' %StyleTableListHistogramWidth% ';
    //kvuli TAGum a dalsim sloupcum, ktere nebudou mit jmeno v INI souboru,
    //nemuzeme rovnou zapsat jmeno promenne, ale musime axistenci zkontrolovat
    s1:=GlobalVarGet('ColumnHeader'+sCol);
    if(s1='ColumnHeader'+sCol) then begin
      s1:=ReplaceString(sCol,'tag-','');
    end;
    Result:=Result+'    <th style="%StyleTableListTh%'+sWidth+'"';
    //if(sCol='Histogram') then Result:=Result+' WIDTH="150"';
    if(s1='') then s1:='&nbsp;';
    Result:=Result+'>'+s1+'</th>'+CRLF;
  end;
  Result:=Result+'  </tr>'+CRLF+'</thead><tbody>'+CRLF;
end;

function AndIfNeed(str:string):string;
begin
  Result:='';
  if(str<>'') then Result:=' AND ';
end;

function CommaIfNeed(str:string):string;
begin
  Result:='';
  if(str<>'') then Result:=', ';
end;

function MakeTagWhereItem(flags:REC_COLUMN_FLAGS):string;
var s:string;
begin
  //zkonstruuje frazi WHERE podle pozadovaneho operatoru
  //hodnotu tagu musime prevest z ANSI na UTF-8
  //!!!!! ale knigovna neni kompatibilni s unicodestring, takze stale ocekava
  //!!!!! ansi, proto nelze prevadet na UTF
  s:={AnsiToUtf}(flags.sValue);
  case flags.sOperator of
  'IN':      Result:='value IN ('+s+')';
  'NOT IN':  Result:='value NOT IN ('+s+')';
  'LIKE':    Result:='value LIKE '+s;
  'BETWEEN': Result:='value BETWEEN '+ReplaceString(s,',',' AND ');
  else Result:='value'+flags.sOperator+s;
  end;
end;

function MakeWhereItem(flags:REC_COLUMN_FLAGS):string;
var s:string;
begin
  //zkonstruuje frazi WHERE podle pozadovaneho operatoru
  //hodnotu tagu musime prevest z ANSI na UTF-8
  //!!!!! ale knigovna neni kompatibilni s unicodestring, takze stale ocekava
  //!!!!! ansi, proto nelze prevadet na UTF
  s:={AnsiToUtf}(flags.sValue);
  case flags.sOperator of
  'IN':      Result:=' IN ('+s+')';
  'NOT IN':  Result:=' NOT IN ('+s+')';
  'LIKE':    Result:=' LIKE '+s;
  'BETWEEN': Result:=' BETWEEN '+ReplaceString(s,',',' AND ');
  else Result:=flags.sOperator+s;
  end;
end;

function TableListMakeSqlFromParams(aPar:TStringList;var idxCount:integer):string;
var bTag,bFoundOnly:boolean;
    sSelect,sFrom,sWhere,sGroup,sOrder,sCol,sTag,sJoin,sTagID,sTagValue,sStatorTable:string;
    s,s1,s2,sColOrder,sColGroup,sColCount:string;
    flags:REC_COLUMN_FLAGS;
begin
  //sestavime jednotlive casti SQL dotazu podle pozadovanych sloupcu
  bTag:=false;
  sSelect:='';
  sFrom:='geocache gc';
  sWhere:='';
  sGroup:='';
  sOrder:='';
  sTag:='';
  sJoin:='';
  sStatorTable:='';
  //projdeme pozadovane sloupce a doplnime sSelect, sFrom a sWhere
  //pokud hledame jen mezi nalezenymi
  bFoundOnly:=(aPar.Values['FoundOnly']='Yes');
  if(bFoundOnly) then begin
    //sWhere:='gc.dtfound>0';
    sFrom:='temp.Stator st, '+sFrom;
    sWhere:='gc.id=st.id';
  end;
  //if(bFoundOnly) then sWhere:='gc.dtfound>0';
  s:=aPar.Values['Columns'];
  idxCount:=0;  //sem budeme pocitat poradi datoveho sloupce v selectu, z ktereho brat pocet pro soucet, procentra a histogram
  while (s<>'') do begin
    bTag:=false;
    sColOrder:='';   
    s1:=Fetch(s,'|');
    ParseColumnFlags(s1,flags); //parametry tohoto sloupce
    sCol:=flags.sName;
//Log('FFFFF sloupec='+sCol);
    if(Pos('tag-',sCol)=1) then begin
      bTag:=true;
      Fetch(sCol,'-');        //nechame jen kategorii tagu
      //pokud je pozadovana i nejaka hodnota tagu, zjistime ji
      sTagValue:=flags.sValue;
      sTagID:='T'+IntToStr(Length(s));

      if(flags.sOperator='') then begin
        //##s2:='IFNULL((SELECT value FROM geotagvalue WHERE key=g'+sTagID+'.ptrvalue),'''') '
        //##  +'val'+sTagID; //takto se sloupec bude jmenovat
        s2:='(SELECT value FROM geotagvalue WHERE key=g'+sTagID+'.ptrvalue) '
          +'val'+sTagID; //takto se sloupec bude jmenovat
      end
      else begin
      //Log('AAAAAA test: '+MakeTagWhereItem(flags));
        s2:='(SELECT value FROM geotagvalue '
          +'WHERE key=g'+sTagID+'.ptrvalue AND '+MakeTagWhereItem(flags)+' COLLATE NOCASE) '
          +'val'+sTagID; //takto se sloupec bude jmenovat
        sWhere:=sWhere+AndIfNeed(sWhere)+'val'+sTagID+' IS NOT NULL';
      end;
       //jak tady obslouzit bCounting?  ####
      sSelect:=sSelect+CommaIfNeed(sSelect)+s2;
      //nutnost/nepovinnost nalezeni kese uz je obslouzena pridanim tabulky 
      //temp.Stator a podminkou na shodu ID
      if(flags.bTagMustExist) then s2:='JOIN '
      else s2:='LEFT OUTER JOIN ';
      sJoin:=sJoin+CRLF+s2+'geotag g'+sTagID+' ON '
        +'(gc.id=g'+sTagID+'.id and g'+sTagID+'.ptrkat IN (SELECT key FROM geotagcategory '
        +'  WHERE value=:TagCat_'+sTagID+' COLLATE NOCASE))'+CRLF;
      //hodnotu kategorie musime prevest z ANSI na UTF-8
      //Log(':TagCat_'+sTagID+'='+AnsiToUtf(sCol));
      Geoget_DB.AddParamText(':TagCat_'+sTagID,AnsiToUtf(sCol));
//Log('DDDDDD sCol='+sCol+', sSort='+flags.sSort+', bNumeric='+IntToStr(flags.bNumeric as integer));
      if(flags.sSort<>'') then begin
        if(flags.bNumeric) then sColOrder:='(0.+val'+sTagID+')'
        else sColOrder:='val'+sTagID;
      end;
      if(flags.bCounting) then sColGroup:='val'+sTagID;
      if(flags.bCounting) then sColCount:='(SELECT value FROM geotagvalue WHERE key=g'+sTagID+'.ptrvalue)';
    end
    else begin
      case sCol of
      'ID':             begin
                          sCol:='gc.id';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.id'+MakeWhereItem(flags);
                        end;
      'Type icon',
      'Type icon, text',
      'Type':           begin
                          sCol:='gc.cachetype';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.cachetype'+MakeWhereItem(flags);
                        end;
      'Name':           begin
                          sColOrder:='gc.Name';
                          sCol:='CASE (gc.cachestatus) when "0" then "+"||gc.name when "1" then "-"||gc.name when "2" then "#"||gc.name else "?"||gc.name END';
						  sColOrder:='gc.Name';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.name'+MakeWhereItem(flags);
                        end;  
      'Size icon',
      'Size icon, text',
      'Size':           begin
                          sCol:='gc.cachesize';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.cachesize'+MakeWhereItem(flags);
                        end;  
      'Size as num':    begin 
                          sCol:='CASE (gc.cachesize) '
                             +'WHEN "Not chosen" THEN "0 Not chosen" '
                             +'WHEN "Micro" THEN "1 Micro" '
                             +'WHEN "Small" THEN "2 Small" '
                             +'WHEN "Regular" THEN "3 Regular" '
                             +'WHEN "Large" THEN "4 Large" '
                             +'WHEN "Virtual" THEN "5 Virtual" '
                             +'WHEN "Other" THEN "6 Other" '
                             +'ELSE  "7 Unknown" '
                             +'END SizeNum';
                          if(flags.bCounting) then sColCount:='gc.cachesize';
                          if(flags.bCounting) then sColGroup:='gc.cachesize';
                          if(flags.sSort<>'') then sColOrder:='SizeNum';   
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.cachesize'+MakeWhereItem(flags);
                       end; 
      'Difficulty icon',
      'Difficulty icon, text',
      'Difficulty':     begin
                          sCol:='gc.difficulty';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.difficulty'+MakeWhereItem(flags);
                        end;  
      'Terrain icon',
      'Terrain icon, text',
      'Terrain':        begin 
                          sCol:='gc.terrain';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.terrain'+MakeWhereItem(flags);
                        end;  
      //'IDTag':        sCol:=',gc.';
      'Status icon',
      'Status icon, text',
      'Status':         begin
                          sCol:='gc.cachestatus';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.cachestatus'+MakeWhereItem(flags);
                        end;  
      'Author':         begin
                          sCol:='gc.author';
                          if(flags.sOperator<>'') then 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.author'+MakeWhereItem(flags);
                        end;  
      'DateCreate':     begin
                          sCol:='gc.dthidden';
                          if(flags.sOperator<>'') then
                            //tady fakt nvim, jak by mela hodnota vypadat, aby to fungovalo 
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.author'+MakeWhereItem(flags);
                        end;  
      'Distance':       begin
                          sCol:='st.distance';
                          if(Pos('Stator st',sFrom)<1) then begin
                            sFrom:=sFrom+', temp.Stator st';
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.id=st.id';
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'st.distance'+MakeWhereItem(flags);
                          end;
                        end;
      'Direction',
      'Direction icon': begin
                          sCol:='st.angle';
                          if(Pos('Stator st',sFrom)<1) then begin
                            sFrom:=sFrom+', temp.Stator st';
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.id=st.id';
                          end;
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'st.angle'+MakeWhereItem(flags);
                        end;
      'Coordinates':    sCol:='gc.x||";"||gc.y';
      'Country flag',
      'Country flag, text',
      'Country':        begin
                          sCol:='gc.country';
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.country'+MakeWhereItem(flags);
                        end;
      'Cache info':     sCol:='gc.id||";"||gc.cachetype||";"||gc.cachestatus||";"||gc.country||";"||gc.name';
      'State':          begin
                          sCol:='gc.state';
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.state'+MakeWhereItem(flags);
                        end;
      'Found date':     begin
                          sCol:='gc.dtfound';
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'gc.dtfound'+MakeWhereItem(flags);
                        end;
      'Found time':     sCol:='substr("000"||gc.dtfoundtime,-4,2)||":"||substr("000"||gc.dtfoundtime,-2,2)';
      'Found DateTime': sCol:='gc.dtfound*10000+gc.dtfoundtime';
      'Date of last found': sCol:='gc.dtlast';
      'Count':          begin
                          sCol:='count(%COUNT%)';
                          if(flags.sSort<>'') then sColOrder:='count(%COUNT%)';
                          {
                          if(iOrder=0) then sColOrder:='ORDER BY count(%COUNT%) desc'; //######
                          //pokud je sloupec zaroven poctem, nastavime jinak, aby nedoslo k "cykleni"
                          if(iCount=0) then begin
                            sCol:='count(gc.id)';
                            sCount:='count(gc.id)';
                            if(iOrder=0) then sOrder:='ORDER BY count(gc.id) desc'; //#####
                          end;
                          }
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'count(%COUNT%)'+MakeWhereItem(flags);
                        end;
      'Order finding',  //toto je spatne a nechavan to jen z historickych duvodu
                        //aby neprestaly chodit stare selecty
      'Order by finds':  begin
                          if(bFoundOnly) then begin
                            if(Pos('Stator st',sFrom)<1) then begin
                              sWhere:=sWhere+AndIfNeed(sWhere)+'gc.id=st.id';
                            end;
                          end;
                          //pokud neni bFoundOnly, tak neni pripojena tabulka st!!!!!!
                          //ale jen pokud jiz tabulka byla pridana
                          if((not bFoundOnly) and (Pos('temp.Stator st',sJoin)<1)) then begin
                            sJoin:=sJoin+CRLF+'LEFT OUTER JOIN temp.Stator st ON (gc.id=st.id)';
                          end;
                          sCol:='st.rowid';
                          if(flags.sOperator<>'') then
                            sWhere:=sWhere+AndIfNeed(sWhere)+'st.rowid'+MakeWhereItem(flags);
                        end;
      'Milestone':      begin
                          //sFrom:=sFrom+', temp.Stator st';
                          //sWhere:=sWhere+AndIfNeed(sWhere)+'gc.id=st.id';
                          sCol:='distinct st.rowid Milestones';
                          if(flags.sSort<>'') then sColOrder:='Milestones';
                        end;
      'Milestone, interval': sCol:='gc.dtfound';
      'Order row':      sCol:='"?"';       //pri SQL ignorujeme
      'Percent':        sCol:='"?"';      //pri SQL ignorujeme
      'Percent explicit': sCol:='"?"';    //pri SQL ignorujeme
      'Histogram':      sCol:='"?"';     //pri SQL ignorujeme
      else begin Log(_('ERR: unknown column name:')+' '+sCol); continue; end;
      end;
      //doplnime podle sloupce podminku, pokud je potreba a neni to 
    end;
    if(flags.bCounting) then begin
      if(sColGroup='') then sColGroup:=sCol;
      sGroup:='GROUP BY '+sColGroup;
      if(sColCount='') then sColCount:=sCol;
    end;
    if(sCol<>'') then begin
      if(not bTag) then sSelect:=sSelect+CommaIfNeed(sSelect)+sCol;
//Log('HHHHH sSelect='+sSelect);
      if(flags.sSort<>'') then begin
        if(sColOrder='') then begin
          if(not flags.bNumeric) then sColOrder:=sCol
          else sColOrder:='(0.+'+sCol+')';
        end;
        if(sOrder='') then sOrder:='ORDER BY '
        else sOrder:=sOrder+', ';
        sOrder:=sOrder+sColOrder+' '+flags.sSort;
      end;
    end;
  end;
  //protoze pocet mohl byt pred sloupcem, nahradime ted spravny sloupec pro pocet
  //???? co kdyz neni zadny sloupec, z ktereho by se pocet bral?
  //     (to by bylo v pripade, kdy sloupec pro vypocet je jen z databaze, ale
  //     nikam se do vystupni tabulky nevypisuje)
  if(Pos('%COUNT%',sSelect)>0) and (sColCount='') then begin
    Log(_('ERR: Index of SQL column for Count not set'));
    exit;
  end;
  {
  if(Pos('%GROUP%',sSelect)>0) and (sColGroup='') then begin
    Log(_('ERR: Index of SQL column for Group not set'));
    exit;
  end;
  }
  sSelect:=ReplaceString(sSelect,'%COUNT%',sColCount);
  sGroup:=ReplaceString(sGroup,'%COUNT%',sColCount);
  sOrder:=ReplaceString(sOrder,'%COUNT%',sColCount);
  //pokud bylo vice tagu, uzavreme zavorku
  Result:='SELECT '+sSelect+CRLF+' FROM '+sFrom+sTag+sJoin;
  if(sWhere<>'') then Result:=Result+CRLF+'WHERE '+sWhere;
  Result:=GlobalVarReplace(Result+CRLF+' '+sGroup+CRLF+' '+sOrder);
  Log('SQL= '+Result);
end;

function MakeHtmlCacheInfo(sId,sType,sName,sCountry,sStatus:string):string;
var s1,s2:string;
begin
  //pokud je sCountry='', nekresli se vlajecka
  //pokud je iStatus jiny nez '1'/'2', nemeni se barva a styl pisma podle stavu kese
  case sStatus of
  '1': begin s2:='</span>'; s1:='<span style="text-decoration:line-through; color:#%ColorCacheDisabledText%">'; end;
  '2': begin s2:='</span>'; s1:='<span style="text-decoration:line-through; color:#%ColorCacheArchivedText%">'; end;
  end;
  Result:='<img src="%IconCacheTypeSmall'+sType+'%" /> '
      +'<a href="http://coord.info/'+sId+'"';
  if(sStatus<>'') then Result:=Result+' target="_blank" title="%TextCacheStatus'+sStatus+'%"';
  Result:=Result+'>'+s1+sId+s2+'</a> - '+s1+sName+s2;

  if(sCountry<>'') then Result:=Result
    +' <img src="%IconCountry'+sCountry+'%" title="'+sCountry+'" />';
end;

function MakeHtmlCacheInfoFromId(sId:string):string;
var bod:TGeo;
begin
  Result:='';
  bod:=TGeo.Create();
  try
    bod.LoadByGc(sId);
    //bod.id:=sId; //toto nestaci, musi byt LoadByGc()
    Result:=MakeHtmlCacheInfo(sId,bod.CacheType,UtfToAnsi(bod.Name),bod.Country,IntToStr(bod.cachestatus));
  finally
    bod.Free();
  end;
end;

function OutputTableListRow(sCols:string; aList:TStringList; iRow,iMin,iMax,iSuma,nCols,idxCount,iHistogramWidth:integer):string;
var sCol,s1,s2,sAlign,sOdd,sPomId,sPomName,sPomType,sPomStatus,sPomCountry:string;
    i,j,iCol:integer;
    f,fX1,fY1,fU,fLength:extended;
    dt:TDateTime;
    flags:REC_COLUMN_FLAGS;
begin
  sOdd:='';
  if((iRow Mod 2)=0) then sOdd:='Odd';
  Result:='  <tr style="%StyleTableListTr'+sOdd+'%">'+CRLF;
  iCol:=0; //zaciname sloupcem 0
  while (sCols<>'') do begin
    s1:=Fetch(sCols,'|');
    ParseColumnFlags(s1,flags);
    if(flags.sAlign='hide') then begin Inc(iCol); continue; end; //preskocime neviditelne sloupce
    sCol:=flags.sName; //odpareme pripadny komentar
    if(flags.sAlign='') then flags.sAlign:=LowerCase(SetDefaultColumnAlign(flags.sName,false));
    sAlign:=flags.sAlign;
    Result:=Result+'    <td style="%StyleTableListTd'+sOdd+'%';
    //zpracujeme
    s2:=UtfToAnsi(aList[iRow*nCols+iCol]);
    //nektere sloupce jsou pocitane, nejsou v datech
   case UpperCase(sCol) of
    'ID':                       s2:='<a href="http://coord.info/'+s2+'" target="_blank">'+s2+'</a>';
    'TYPE':                     ;
    'TYPE ICON':                s2:='<img src="%IconCacheTypeSmall'+s2+'%" />'; 
    'TYPE ICON, TEXT':          s2:='<img src="%IconCacheTypeSmall'+s2+'%" />&nbsp;&nbsp;'+s2;
    'NAME':                     begin
                                  //prvni znak indikuje status kese, pokud je archivovana, piseme preskrtnute
                                  case s2[1] of
                                  '#': s2:='<span style="text-decoration:line-through; color:#%ColorCacheArchivedText%">'+copy(s2,2,length(s2)-1)+'</span>';
                                  '-': s2:='<span style="text-decoration:line-through; color:#%ColorCacheDisabledText%">'+copy(s2,2,length(s2)-1)+'</span>';
                                  '+', //dostupna
                                  '?': //neznamo
                                       s2:=copy(s2,2,length(s2)-1);
                                  //else;
                                  end;
                                end;
    'SIZE':                     ;
    'SIZE ICON':                s2:='<img src="%IconCacheSize'+s2+'%" />'; 
    'SIZE ICON, TEXT':          s2:='<img src="%IconCacheSize'+s2+'%" />&nbsp;&nbsp;'+s2; 
    'SIZE AS NUM': //tento sloupec se pouziva jako skryty jen pro razeni, ale budiz, kdyz nekdo chce ...
                                begin 
                                  sAlign:='left';
                                  Fetch(s2,' '); 
                                end;
    'DIFFICULTY':               ;
    'DIFFICULTY ICON':          s2:='<img src="%IconCacheDiff'+s2+'%" />'; 
    'DIFFICULTY ICON, TEXT':    s2:='<img src="%IconCacheDiff'+s2+'%" />&nbsp;&nbsp;'+s2; 
    'TERRAIN':                  ;
    'TERRAIN ICON':             s2:='<img src="%IconCacheTerr'+s2+'%" />'; 
    'TERRAIN ICON, TEXT':       s2:='<img src="%IconCacheTerr'+s2+'%" />&nbsp;&nbsp;'+s2; 
    //'IDTAG':                  ;
    'STATUS':                   s2:='%TextCacheStatus'+s2+'%'; 
    'STATUS ICON':              s2:='<img src="%IconCacheStatus'+s2+'%" title="%TextCacheStatus'+s2+'%"/>';
    'STATUS ICON, TEXT':        s2:='<img src="%IconCacheStatus'+s2+'%" />&nbsp;&nbsp;'+'%TextCacheStatus'+s2+'%';
    'AUTHOR':                   ;
    'DATECREATE':               begin 
                                  s2:=FormatDateTime('d.m.yyyy',IntToDate(s2));
                                  if(s2='30.12.1899') then s2:='';
                                end;
    'DISTANCE':                 begin 
                                  //fX1:=StrToFloat(Fetch(s2,';'));
                                  //fY1:=StrToFloat(s2);
                                  //fLength:=PointDistanceEx(fHomeX,fHomeY,fX1,fY1,fU);
                                  if(s2<>'') then begin //pokud je nejaka hodnota, prevedeme ji na float
                                    fLength:=StrToFloat(s2);
                                    //prevedeme na pozadovane jednotky
                                    fLength:=ConvertFromKm(fLength,sUnit);
                                    s2:=FormatFloatDot('0.00',fLength);
                                  end;
                                end;
    'DIRECTION':                begin 
                                  //fX1:=StrToFloat(Fetch(s2,';'));
                                  //fY1:=StrToFloat(s2);
                                  //fLength:=PointDistanceEx(fHomeX,fHomeY,fX1,fY1,fU);
                                  if(s2<>'') then begin //pokud je nejaka hodnota, prevedeme ji na float
                                    fU:=StrToFloat(s2);
                                    s2:=IntToStr(Round(fU))+' ';
                                  end;
                                end;
    'DIRECTION ICON':           begin 
                                  //fX1:=StrToFloat(Fetch(s2,';'));
                                  //fY1:=StrToFloat(s2);
                                  //fLength:=PointDistanceEx(fHomeX,fHomeY,fX1,fY1,fU);
                                  if(s2<>'') then begin //pokud je nejaka hodnota, prevedeme ji na float
                                    fU:=StrToFloat(s2);
                                    i:=45*round((fU+(45/2))/45);
                                    s2:='<img src="%IconDirection'+IntToStr(i)+'%"'
                                      //+' title="L='+FormatFloatDot('0.00',fLength)+'; U='+FloatToStr(fU)+'"'
                                      +' />';
                                  end;
                                end;
    'COORDINATES':              begin 
                                  fX1:=StrToFloat(Fetch(s2,';'));
                                  fY1:=StrToFloat(s2);
                                  s2:=FormatCoordNum(fX1,fY1);
                                end;
    'COUNTRY':                  ;
    'CACHE INFO':               begin
                                  sPomId:=Fetch(s2,';');
                                  sPomType:=Fetch(s2,';');
                                  sPomStatus:=Fetch(s2,';');
                                  sPomCountry:=Fetch(s2,';');
                                  sPomName:=s2;   
                                  s2:=MakeHtmlCacheInfo(sPomId,sPomType,sPomName,sPomCountry,sPomStatus);
                                end;  
    'COUNTRY FLAG':             s2:='<img src="%IconCountry'+s2+'%" title="'+s2+'" />';
    'COUNTRY FLAG, TEXT':       s2:='<img src="%IconCountry'+s2+'%" />&nbsp;&nbsp;'+s2; 
    'STATE':                    ;
    'FOUND DATE':               begin
                                  s2:=FormatDateTime('d.m.yyyy',IntToDate(s2));
                                  if(s2='30.12.1899') then s2:='';
                                end;
    'FOUND TIME':               begin
                                  if(Pos(':',s2)<1) then begin
                                    s2:=Copy('0000'+s2,length(s2)+4-3,length(s2)+4);
                                    s2:=Copy(s2,1,2)+':'+Copy(s2,3,length(s2));
                                  end;
                                end;
    'FOUND DATETIME':           begin 
                                  s2:=FormatDateTime('d.m.yyyy',IntToDate(Copy(s2,1,8)))
                                    +' '+Copy(s2,9,2)+':'+Copy(s2,11,2);
                                  if(s2='30.12.1899 :') then s2:='';
                                end;
    'DATE OF LAST FOUND':       begin 
                                  s2:=FormatDateTime('d.m.yyyy',IntToDate(s2));
                                  if(s2='30.12.1899') then s2:='';
                                end;
    'COUNT':                    ;
    'ORDER FINDING':            ; //toto je spatne a nechavan to jen z historickych duvodu
                                  //aby neprestaly chodit stare selecty
    'ORDER BY FINDS':           ;
    'ORDER ROW':                s2:=IntToStr(iRow+1);
    'MILESTONE':                ;
    'MILESTONE, INTERVAL':      begin 
                                  //prevedeme na datum a cas a odecteme posledni hodnotu,
                                  //tim ziskame pocet dni od posledniho milniku
                                  dt:=IntToDate(s2);
                                  if(dtMilestoneLastFound=0) then s2:=''
                                  else s2:=IntToStr(Trunc(dt-dtMilestoneLastFound));
                                  dtMilestoneLastFound:=dt;
                                end;
    'PERCENT':                  begin
                                  if(idxCount>=0) then begin
                                    i:=StrToInt(aList[iRow*nCols+idxCount]);
                                    f:=100*i;
                                    if(iSuma=0) then s2:='0.00'
                                    else s2:=FormatFloatDot('0.00',f/iSuma);
                                  end else s2:='';
                                end;
    'PERCENT EXPLICIT':         begin //procenta primo jako vysledek ve sloupci SQL
                                  s2:=ReplaceString(s2,'.',','); //at mame oddelovac vzdy stejny
                                end;
    'HISTOGRAM':                begin
                                  if(idxCount>=0) then begin
                                    //sirka histogramu je iHistogramWidth
                                    i:=StrToInt(aList[iRow*nCols+idxCount]);
                                    if(iMax=0) then j:=0
                                    else //j:=Trunc(0.9+iHistogramWidth*i/iMax);
                                         j:=Trunc(0.9+100*i/iMax);
                                    s2:='<div style="width:'+IntToStr(j)
                                      //+'px; background-color:#'
                                      +'%; background-color:#'
                                      +ScaleColor(i,0,iMax,GlobalVarGet('ColorTableScaleMinBkg'),GlobalVarGet('ColorTableScaleAvgBkg'),GlobalVarGet('ColorTableScaleMaxBkg'))
                                      +';%StyleTableListHistogram%">&nbsp;</div>';
                                  end else s2:='';
                                end;
    'HISTOGRAM EXPLICIT':       begin //procenta histogramu primo jako vysledek ve sloupci SQL
                                  s2:=ReplaceString(s2,',','.');
                                  j:=Round(StrToFloat(s2));
                                  s2:='<div style="width:'+IntToStr(j)
                                    //+'px; background-color:#'
                                    +'%; background-color:#'
                                    +ScaleColor(i,0,iMax,GlobalVarGet('ColorTableScaleMinBkg'),GlobalVarGet('ColorTableScaleAvgBkg'),GlobalVarGet('ColorTableScaleMaxBkg'))
                                    +';%StyleTableListHistogram%">&nbsp;</div>';
                                end;
    //protoze funkce muze byt volana po obecnem SQL, muze byt sloupec oznacen
    //jakkoli a muze mit skutecne libovolny obsah. Proto nejde hlasit chybu.
    //else if(Pos('tag-',sCol)<>1) then Log(_('ERR: unknown column name:')+' '+sCol);
    end;
    Inc(iCol); //budem nacitat hodnotu z dalsiho sloupce
    if(sAlign<>'') then sAlign:=' text-align:'+LowerCase(sAlign)+';';
    Result:=Result+sAlign+'">'+s2+'</td>'+CRLF;
  end;
  Result:=Result+'  </tr>'+CRLF;
end;

function OutputTableList(sCols:string; aData:TStringList; iRowFirst,nRowCount,idxCount:integer; var nRest:integer):string;
var s,s2:string;
    i,n,nCols,nRows,iMin,iMax,iSuma,iHistogramWidth:integer;
begin
  //spocitame pocet sloupcu a poradi sloupce pro zjisteni poctu
  s:=sCols;
  nCols:=0;
  n:=-1;
  nRest:=0;
  while(s<>'') do begin
    s2:=Fetch(s,'|');
    Inc(nCols);
  end;
  Result:=OutputTableListHeader(sCols);
  if(aData<>nil) then begin
    s:=GlobalVarGet('StyleTableListHistogramWidth');  //zjistime sirku histogramu pro vypocet
    s:=RegexSubstitute('width:\s*(\d+)',s,'$1');
    if(s='') then begin
      Log(_('WARN: Cannot read width of histogram from')+' Styles.ini, TableListHistogramWidth');
      iHistogramWidth:=150;
    end
    else iHistogramWidth:=StrToInt(s);
    //spocitame minimum, maximum a soucet
    nRows:=aData.Count/nCols;   //pocet datovych radku
    iMin:=1; iMax:=1; iSuma:=0;
    if(idxCount>=0) then begin
      iMin:=999999; iMax:=-999999;
      for i:=0 to nRows-1 do begin
        n:=StrToInt(aData[i*nCols+idxCount]);
        iSuma:=iSuma+n;
        if(n<iMin) then iMin:=n;
        if(n>iMax) then iMax:=n;
      end;
    end;

    //spocitame index posledniho vypisovaneho radku
    if(iRowFirst<0) then iRowFirst:=0;
    if(nRowCount<0) then nRowCount:=nRows;
//Log('SSSSSSSS idxCount='+IntToStr(idxCount)+', nCols='+IntToStr(nCols)
//  +', nRows='+IntToStr(nRows)
//  +', iMin='+IntToStr(iMin)+', iMax='+IntToStr(iMax)
//  +', iSuma='+IntToStr(iSuma)
//  +', nRowCount='+IntToStr(nRowCount)
//  );
  //Log('SSSS iRowFirst='+IntToStr(iRowFirst)+', nRowCount='+IntToStr(nRowCount)+', nRows='+IntToStr(nRows));
    if(iRowFirst<nRows) then begin
      n:=nRowCount;
      if(n+iRowFirst>=nRows) then n:=nRows-iRowFirst;
  //Log('SSSS iRowFirst='+IntToStr(iRowFirst)+', nRowCount='+IntToStr(nRowCount)+', n='+IntToStr(n)+', nRows='+IntToStr(nRows));
      for i:=0 to n-1 do begin
        Result:=Result+OutputTableListRow(sCols,aData,i+iRowFirst,iMin,iMax,iSuma,nCols,idxCount,iHistogramWidth);
      end;
      nRest:=nRows-n-iRowFirst;
    end;
  end;
  Result:=Result+'</tbody></table>';
end;

function OutputChartPieUtf(aValues:TStringList;iHeightGr:integer):string;
var i,iVal,iMax,iSuma,iWidthGr:integer;
    sL,sC,sLabels,sColors,sValues:string;
begin
  //vystup z funkce je vzdy v kodovani UTF, protoze Google API to tak potrebuje
  //v datech je pro kazdou hodnotu jeste Label, za ktery se automaticky
  //do zavorky prida ciselna hodnota a procenta

  // rozmry a parametry grafu
  iWidthGr:=iWidth;
  // test na limitn velikosti grafu v bodech
  if (iWidthGr*iHeightGr) > MAX_GraphAreaSize then iWidthGr:=Trunc(MAX_GraphAreaSize/iHeightGr);

  //napred zjistime celkovy soucet hodnot
  iSuma:=0;
  iMax:=StrToInt(aValues[0]);
  for i:=0 to aValues.Count-1 do begin
    iVal:=StrToInt(aValues[i]);
    iSuma:=iSuma+iVal;
    if(iVal>iMax) then iMax:=iVal;
    Inc(i); //preskocime label
  end;
  for i:=0 to aValues.Count-1 do begin
    iVal:=StrToInt(aValues[i]);
    sL:=aValues[i+1]+' - '+IntToStr(iVal)+' ('
       +FormatFloatDot('0.0',(100.0*iVal/iSuma))+' %)';
    if(i mod 4)=0 then sC:=GlobalVarGet('ColorTableScaleAvgBkg') else sC:=GlobalVarGet('ColorTableScaleMinBkg');
    if(iVal=iMax) then sC:=GlobalVarGet('ColorTableScaleMaxBkg');
    if sLabels='' then sLabels:=EncodeGoogleLabel(sL) else sLabels:=sLabels+'|'+EncodeGoogleLabel(sL);
    if sValues='' then sValues:=IntToStr(iVal) else sValues:=sValues+','+IntToStr(iVal);
    if sColors='' then sColors:=sC else sColors:=sColors+','+sC;
    Inc(i); //preskocime label
  end;
  // vstup do grafu
  Result:='<img src="http://chart.googleapis.com/chart?cht=p3'
    +'&chs='+IntToStr(iWidth)+'x'+IntToStr(iHeightGr)
    +'&chf=bg,s,'+GlobalVarGet('ColorMainBkg')
    +'&chco='+sColors
    +'&chxt=x'
    +'&chxs=0,'+GlobalVarGet('ColorMainText')
    +'&chds=a'
    +'&chd=t:'+sValues
    +'&chl='+sLabels
    +'" />';
  //sLabels uz je prevedeno do UTF a entit, neprevadet znovu
  //Result:=AnsiToUtf(GlobalVarReplace(Result));
end;

function OutputChartBarUtf(aValues:TStringList; iHeightGr:integer; var iGraphWd:integer; iRange:integer):string;
var i,ii,iMax,iVal,iCount:integer;
    sVal,sData1,sData2,sD1,sD2,sMark,sM,sLabAx,sLabel,sLabVal:string;
    iBarWd,iBarSp:integer;
begin
  //vystup z funkce je vzdy v kodovani UTF, protoze Google API to tak potrebuje
  iCount:=aValues.Count/3;
  iBarSp:=4;                                  // mezera mezi bary histogramu
  iBarWd:=Round((iGraphWd-6)/iCount)-iBarSp;  // ka baru histogramu
  iGraphWd:=(iBarWd+iBarSp)*iCount+6;         // pepotan ka graf (vrac se v promnn do volajc procedury)

  // zjistme maximln hodnotu dat, kterou pak ve grafu zvraznme
  for i:=0 to aValues.Count-1 do begin
    iVal:=StrToInt(aValues[i]);
    if(iVal>iMax) then iMax:=iVal;
    Inc(i); Inc(i);                 //preskocime label na ose a label na baru
  end;
  // hlavn smyka hodnot, label a popisk grafu
  ii:=0;
  for i:=0 to aValues.Count-1 do begin
    // ve stringlistu jsou za sebou: 1.hodnota, 2.popis osy, 3.popis baru
    sVal:=aValues[i];
    sLabAx:=EncodeGoogleLabel(aValues[i+1]);
    sLabVal:=EncodeGoogleLabel(aValues[i+2]);
    iVal:=StrToInt(sVal);
    if (iVal=iMax) then begin
      sD1:='0'; 
      sD2:=sVal;
    end else begin
      sD1:=sVal; 
      sD2:='0';
    end;
    if(sData1='') then sData1:=sD1 else sData1:=sData1+','+sD1;
    if(sData2='') then sData2:=sD2 else sData2:=sData2+','+sD2;
    sLabel:=sLabel+'|'+sLabAx;
    sM:='t'+sLabVal+','+GlobalVarGet('ColorMainText')+',0,'+IntToStr(ii)+',10';
    if(iVal=iMax) then sM:='t'+sLabVal+','+GlobalVarGet('ColorMainText')+',1,'+IntToStr(ii)+',10';
    if(sMark='') then sMark:=sM else sMark:=sMark+'|'+sM;
    Inc(i); Inc(i);                 //preskocime na dal zznam
    Inc(ii);
  end;
  if iRange=0 then iRange:=iMax;    //pokud nen zadn range extern, nastav se podle maximln hodnoty lokln
  Result:='<img src="http://chart.googleapis.com/chart?cht=bvs'
       +'&chs='+IntToStr(iGraphWd)+'x'+IntToStr(iHeightGr)
       +'&chbh='+IntToStr(iBarWd)+','+IntToStr(iBarSp)
       +'&chds=0,'+IntToStr(Round(iRange*1.2))+',0,'+IntToStr(Round(iRange*1.2)) //data scale, rozsah v ose y trochu vetsi nez nejvyssi hodnota, aby se vely popisky
       +'&chf=bg,s,'+GlobalVarGet('ColorMainBkg')
       +'&chco='+GlobalVarGet('ColorTableScaleAvgBkg')+','+GlobalVarGet('ColorTableScaleMaxBkg')
       +'&chxt=x'
       +'&chd=t:'+sData1+'|'+sData2 
       +'&chxl=0:'+sLabel
       +'&chxs=0,'+GlobalVarGet('ColorMainText')+',10'
       +'&chm='+sMark
       +'" />';
  //sLabels uz je prevedeno do UTF a entit, neprevadet znovu
  //Result:=AnsiToUtf(Result);
end;
